# Copyright (c) 2019, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import awkward


class UprootTransformer:
    def __init__(self, event_iterator):
        self.event_iterator = event_iterator

    @property
    def file_path(self):
        return self.event_iterator.file_path

    @property
    def attr_name_list(self):
        return self.event_iterator.attr_name_list

    @property
    def chunk_size(self):
        return self.event_iterator.chunk_size

    def arrow_table(self):
        for object_array in self.event_iterator.iterate():
            unicode_array = {}
            for key in object_array.keys():
                unicode_array[key.decode('UTF8')] = object_array[key]
            yield awkward.toarrow(awkward.Table(unicode_array))



# Copyright (c) 2019, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from pyarrow.parquet import ParquetWriter
import pyarrow as pa


class TestScratchFileWriter:
    def _test_table(self):
        schema = pa.schema([('field1', pa.int32())])
        return pa.Table.from_arrays([[1]], schema=schema)

    def test_init(self, mocker):
        from servicex.transformer.scratch_file_writer import ScratchFileWriter

        writer = ScratchFileWriter('parquet')
        assert writer.file_path == '/tmp/out'
        assert not writer.scratch_file
        assert writer.file_format == 'parquet'

    def test_open_scratch_file(self, mocker):
        table = self._test_table()

        mock_parquet = mocker.MagicMock(ParquetWriter)
        foo = mocker.patch("pyarrow.parquet.ParquetWriter", return_value=mock_parquet)

        from servicex.transformer.scratch_file_writer import ScratchFileWriter
        writer = ScratchFileWriter('parquet')

        writer.open_scratch_file(table)
        foo.assert_called_with("/tmp/out", table.schema)

    def test_append_table_to_scratch(self, mocker):
        mock_parquet = mocker.MagicMock(ParquetWriter)
        mocker.patch("pyarrow.parquet.ParquetWriter", return_value=mock_parquet)

        table = self._test_table()
        from servicex.transformer.scratch_file_writer import ScratchFileWriter
        writer = ScratchFileWriter('parquet')

        writer.open_scratch_file(table)
        writer.append_table_to_scratch(table)
        mock_parquet.write_table.assert_called()

    def test_close_scratch_file(self, mocker):
        table = self._test_table()

        mock_parquet = mocker.MagicMock(ParquetWriter)
        mocker.patch("pyarrow.parquet.ParquetWriter", return_value=mock_parquet)

        from servicex.transformer.scratch_file_writer import ScratchFileWriter
        writer = ScratchFileWriter('parquet')
        writer.open_scratch_file(table)
        writer.close_scratch_file()
        mock_parquet.close.assert_called()
        assert not writer.scratch_file

    def test_remove_scratch_file(self, mocker):
        mock_remove = mocker.patch("os.remove")
        from servicex.transformer.scratch_file_writer import ScratchFileWriter
        writer = ScratchFileWriter('parquet')
        writer.remove_scratch_file()
        mock_remove.assert_called_with("/tmp/out")



# Copyright (c) 2019, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import ROOT


class XAODEvents:
    def __init__(self, file_path, attr_name_list, chunk_size, event_limit=None):
        self.file_path = file_path
        print("Openning ", file_path)
        self.file_in = ROOT.TFile.Open(file_path)
        print(self.file_in)
        self.tree = ROOT.xAOD.MakeTransientTree(self.file_in)
        self.attr_name_list = attr_name_list
        self.event_limit = event_limit
        self.chunk_size = chunk_size

    def _create_branch_dict(self):
        branches = {}
        for attr_name in self.attr_name_list:
            attr_name = str(attr_name)
            if not attr_name.split('.')[0].strip(' ') in branches:
                branches[attr_name.split('.')[0].strip(' ')] = [
                    attr_name.split('.')[1]]
            else:
                branches[attr_name.split('.')[0].strip(' ')].append(
                    attr_name.split('.')[1])
        return branches

    def _select_branches(self):
        self.branches = self._create_branch_dict()

        self.tree.SetBranchStatus('*', 0)
        self.tree.SetBranchStatus('EventInfo', 1)
        for branch_name in self.branches:
            self.tree.SetBranchStatus(branch_name, 1)

    def get_entry_count(self):
        return self.tree.GetEntries()

    def iterate(self):
        self._select_branches()

        n_entries = self.tree.GetEntries()
        if self.event_limit:
            n_entries = min(n_entries, self.event_limit)
            print("Limiting to the first " + str(n_entries) + " events")

        for j_entry in range(n_entries):
            self.tree.GetEntry(j_entry)
            if j_entry % 1000 == 0:
                print("Processing run #" + str(self.tree.EventInfo.runNumber())
                      + ", event #" + str(self.tree.EventInfo.eventNumber())
                      + " (" + str(
                            round(100.0 * j_entry / n_entries, 2)) + "%)")

            particles = {}
            full_event = {}
            for branch_name in self.branches:
                full_event[branch_name] = []
                particles[branch_name] = getattr(self.tree, branch_name)
                for i in xrange(particles[branch_name].size()):
                    particle = particles[branch_name].at(i)
                    single_particle_attr = {}
                    for a_name in self.branches[branch_name]:
                        single_particle_attr[a_name] = \
                            getattr(particle, a_name.strip('()'))()
                    full_event[branch_name].append(single_particle_attr)

            yield full_event



# Copyright (c) 2019, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from collections import OrderedDict


# noinspection PyClassHasNoInit
class TestXAODTransformer:

    def test_arrow_table(self, mocker):
        import mock
        from servicex.transformer.xaod_transformer import XAODTransformer

        attr_names = [
            "Electrons.pt()", "Electrons.eta()",  "Muons.e()"
        ]

        from servicex.transformer.xaod_events import XAODEvents
        iterator = mock.MagicMock(XAODEvents)
        iterator.iterate = mock.Mock(return_value=iter([
            {'Electrons': [
                {'pt()': 4.0, 'eta()': 5.0},
                {'pt()': 8.0, 'eta()': 10.0}
            ],
                'Muons': [
                    {'e()': 1.0}, {'e()': 2.0}, {'e()': 3.0}
                ]
            },
            {'Electrons': [
                {'pt()': 14.0, 'eta()': 15.0},
                {'pt()': 18.0, 'eta()': 20.0}
            ],
                'Muons': [
                    {'e()': 1.0}, {'e()': 2.0}, {'e()': 3.0}
                ]
            }
        ]))

        iterator.attr_name_list = attr_names
        iterator.get_entry_count = mock.Mock(return_value=1000)
        iterator.chunk_size = 2
        transformer = XAODTransformer(iterator)
        table = transformer.arrow_table().next()

        assert table.column_names == ['Electrons_pt',
                                      'Muons_e',
                                      'Electrons_eta']

        assert len(table) == 2
        assert table.num_rows == 2
        assert table.num_columns == 3
        assert table.shape == (2, 3)
        pydict = table.to_pydict()
        print(pydict)
        assert pydict == OrderedDict([
            ('Electrons_pt', [[4.0, 8.0], [14.0, 18.0]]),
            ('Muons_e', [[1.0, 2.0, 3.0],  [1.0, 2.0, 3.0]]),
            ('Electrons_eta', [[5.0, 10.0], [15.0, 20.0]])
        ])

        iterator.iterate.assert_called()



# Copyright (c) 2019, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import sys
from .messaging import Messaging
from kafka import KafkaProducer


class KafkaMessaging(Messaging):
    def __init__(self, brokers, max_message_size=15):

        print("Max Message size: " + str(max_message_size) + "Mb")
        self.max_message_size = max_message_size

        if not brokers:
            self.brokers = ['servicex-kafka-0.slateci.net:19092',
                            'servicex-kafka-1.slateci.net:19092',
                            'servicex-kafka-2.slateci.net:19092']
        else:
            self.brokers = brokers

        self.producer = None
        print('Configured Kafka backend')

        try:
            self.producer = KafkaProducer(bootstrap_servers=self.brokers,
                                          api_version=(0, 10),
                                          max_request_size=int(max_message_size * 1e6))
            print("Kafka producer created successfully")
        except Exception as ex:
            print("Exception while getting Kafka producer", ex)
            sys.exit(1)

    def publish_message(self, topic_name, key, value_buffer):
        try:
            msg_bytes = value_buffer.to_pybytes()
            self.producer.send(topic_name, key=str(key),
                               value=msg_bytes)
            self.producer.flush()
            print("Message published to ", topic_name, " successfully ",
                  len(msg_bytes))
        except Exception as ex:
            print("Exception in publishing message", ex)
            raise
        return True



# Copyright (c) 2019, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import argparse
import sys
from servicex.transformer.transformer_argument_parser import TransformerArgumentParser


class TestTransformerArgumentParser:
    def test_init(self, mocker):
        argparse.ArgumentParser.exit = mocker.Mock()
        arg_parser = TransformerArgumentParser(description="Test Transformer")
        sys.argv = ["foo", "bar"]
        arg_parser.parse_args()
        argparse.ArgumentParser.exit.assert_called()

    def test_parse(self, mocker):
        arg_parser = TransformerArgumentParser(description="Test Transformer")
        sys.argv = ["foo",
                    "--path", "/foo/bar",
                    "--brokerlist", "kafka1.org",
                    "--topic", "mytopic",
                    "--chunks", "100",
                    "--tree", "Events",
                    "--attrs", "a,b,c",
                    "--limit", "10",
                    '--result-destination', 'kafka',
                    '--result-format', 'arrow',
                    "--max-message-size", "42",
                    '--rabbit-uri', "http://rabbit.org",
                    '--request-id', "123-45-678"
                    ]

        args = arg_parser.parse_args()
        assert args.path == '/foo/bar'
        assert args.brokerlist == "kafka1.org"
        assert args.topic == 'mytopic'
        assert args.chunks == 100
        assert args.tree == 'Events'
        assert args.attr_names == 'a,b,c'
        assert args.limit == 10
        assert args.result_destination == 'kafka'
        assert args.result_format == 'arrow'
        assert args.max_message_size == 42
        assert args.rabbit_uri == "http://rabbit.org"
        assert args.request_id == "123-45-678"

    def test_extract_broker_list(self):
        brokers = TransformerArgumentParser.extract_kafka_brokers("kafka1,kafka2")
        assert len(brokers) == 2
        assert brokers[0] == 'kafka1'
        assert brokers[1] == 'kafka2'

    def test_extract_attr_list(self):
        attrs = TransformerArgumentParser.extract_attr_list("a,b,c")
        assert len(attrs) == 3
        assert attrs == ['a', 'b', 'c']



# Copyright (c) 2019, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import os
import time
import redis
import codecs


class RedisMessaging:
    def __init__(self, host='redis.slateci.net', port=6379):

        self.MAX_MESSAGES_PER_REQUEST = 100000
        if 'MAX_MESSAGES_PER_REQUEST' in os.environ:
            self.MAX_MESSAGES_PER_REQUEST = int(os.environ['MAX_MESSAGES_PER_REQUEST'])
        print("max messages per request:", self.MAX_MESSAGES_PER_REQUEST)
        self.host = host
        self.port = port
        self.client = None

        print('Configured Redis backend')

    def set_redis_client(self):
        self.client = redis.Redis(self.host, self.port, db=0)

    def publish_message(self, topic_name, key, value_buffer):
        # connect if not already done
        try:
            while True:
                if self.client:
                    # print('Redis client connected.')
                    break
                self.set_redis_client()
                print('waiting to connect redis client...')
                time.sleep(60)
        except Exception as ex:
            print("Exception in publishing message:", ex)
            raise
        request_id = 'req_id:' + topic_name
        # check if queue is still full, if yes return false.
        if self.client.xlen(request_id) > self.MAX_MESSAGES_PER_REQUEST:
            return False

        # add message
        self.client.xadd(request_id, {
            'pa': key,
            'data': codecs.encode(value_buffer, 'bz2')
        })
        return True

    def request_status_redis(self, topic_name):
        try:
            topic_name = 'req_id:' + topic_name
            request_info = self.client.xinfo_stream(topic_name)
            print('req_info:', request_info)
        except redis.exceptions.ResponseError as redis_e:
            print(redis_e)



# Copyright (c) 2019, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


# noinspection PyClassHasNoInit,PyMethodMayBeStatic
class TestXAODEvents:
    def _generate_mock_phys_obj(self, mock, mock_map):
        a_key = list(mock_map.keys())[0]
        size_vals = [len(o) for o in mock_map[a_key]]

        obj = mock.Mock()
        obj.size = mock.Mock(side_effect=size_vals)
        obj_at_mock = mock.Mock()
        for key in mock_map.keys():
            flatened = [o2 for o in mock_map[key] for o2 in o]
            add_attr_expr = "type(obj_at_mock).%s = mock.Mock(side_effect=%s)" % \
                (key, str(flatened))
            exec add_attr_expr

        obj.at = mock.Mock(return_value=obj_at_mock)
        return obj

    def test_get_entry_count(self, mocker):
        import mock
        import ROOT

        from servicex.transformer.xaod_events import XAODEvents

        attr_names = [
            "Electrons.pt()", "Electrons.eta()",  "Muons.e()"
        ]

        mock_tree = mock.Mock()
        xaod_mock = mocker.patch.object(ROOT, "xAOD")
        mock_file = mock.Mock()
        mocker.patch.object(ROOT.TFile, 'Open',
                                        return_value=mock_file)
        xaod_mock.MakeTransientTree = mock.Mock(return_value=mock_tree)

        mock_tree.GetEntries = mock.Mock(return_value=2)
        mock_tree.GetEntry = mock.Mock()
        event_iterator = XAODEvents("foo/bar", attr_names, chunk_size=100, event_limit=None)

        assert event_iterator.get_entry_count() == 2
        xaod_mock.MakeTransientTree.assert_called()

    def test_event_iter_no_limit(self, mocker):
        import ROOT
        import mock
        import pytest
        from servicex.transformer.xaod_events import XAODEvents

        attr_names = [
            "Electrons.pt()", "Electrons.eta()",  "Muons.e()"
        ]

        mock_file = mock.Mock()
        open_mock = mocker.patch.object(ROOT.TFile, 'Open',
                                        return_value=mock_file)

        mock_tree = mock.Mock()
        xaod_mock = mocker.patch.object(ROOT, "xAOD")
        xaod_mock.MakeTransientTree = mock.Mock(return_value=mock_tree)

        event_iterator = XAODEvents("foo/bar", attr_names, chunk_size=100, event_limit=None)

        mock_tree.GetEntries = mock.Mock(return_value=2)
        mock_tree.GetEntry = mock.Mock()

        mock_tree.Muons = self._generate_mock_phys_obj(mock, {
            "e": [
                [1, 2, 3],
                [4, 5]
            ]
        })

        mock_tree.Electrons = self._generate_mock_phys_obj(mock, {
            "pt": [
                [4, 8],
                [16, 20, 24]
            ],
            "eta": [
                [5, 10],
                [20, 25, 30]
            ]
        })

        array_gen = event_iterator.iterate()

        array = array_gen.next()
        print(array)
        mock_tree.GetEntry.assert_called_with(0)
        mock_tree.Muons.at.assert_called_with(2)
        mock_tree.Electrons.at.assert_called_with(1)
        open_mock.assert_called_with("foo/bar")

        assert len(array['Muons']) == 3
        assert len(array['Electrons']) == 2
        assert array['Electrons'] == [{'pt()': 4, 'eta()': 5},
                                      {'pt()': 8, 'eta()': 10}]
        assert array['Muons'] == [{'e()': 1}, {'e()': 2}, {'e()': 3}]

        array2 = array_gen.next()
        print(array2)
        mock_tree.GetEntry.assert_called_with(1)
        mock_tree.Muons.at.assert_called_with(1)
        mock_tree.Electrons.at.assert_called_with(2)
        assert len(array2['Muons']) == 2
        assert len(array2['Electrons']) == 3

        assert array2['Electrons'] == [{'pt()': 16, 'eta()': 20},
                                       {'pt()': 20, 'eta()': 25},
                                       {'pt()': 24, 'eta()': 30}]

        assert array2['Muons'] == [{'e()': 4}, {'e()': 5}]

        with pytest.raises(StopIteration):
            array_gen.next()
        xaod_mock.MakeTransientTree.assert_called()

    def test_chunksize_greater_than_events(self, mocker):
        assert True

    def batch_larger_than_max_message_bytes(self, mocker):
        assert True



# Copyright (c) 2019, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from servicex.transformer.servicex_adapter import ServiceXAdapter


class TestServiceXAdapter:

    def test_put_file_complete(self, mocker):
        mock_requests_put = mocker.patch('requests.put')
        adapter = ServiceXAdapter("http://foo.com")
        adapter.put_file_complete("my-root.root", 42, "testing", 1, 2, 3, 4)
        mock_requests_put.assert_called()
        args = mock_requests_put.call_args
        assert args[0][0] == 'http://foo.com/file-complete'
        doc = args[1]['json']
        assert doc['status'] == 'testing'
        assert doc['total-events'] == 3
        assert doc['total-time'] == 2
        assert doc['file-path'] == 'my-root.root'
        assert doc['num-messages'] == 1
        assert doc['file-id'] == 42
        assert doc['avg-rate'] == 1

    def test_post_status_update(self, mocker):
        mock_requests_post = mocker.patch('requests.post')
        adapter = ServiceXAdapter("http://foo.com")
        adapter.post_status_update("testing")
        mock_requests_post.assert_called()
        args = mock_requests_post.call_args
        assert args[0][0] == 'http://foo.com/status'
        doc = args[1]['data']
        assert doc['status'] == 'testing'



# Copyright (c) 2019, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import argparse

default_brokerlist = "servicex-kafka-0.slateci.net:19092, " \
                     "servicex-kafka-1.slateci.net:19092," \
                     "servicex-kafka-2.slateci.net:19092"

default_attr_names = "Events.Electron_pt,Events.Electron_eta,Muon_phi"

# What is the largest message we want to send (in megabytes).
# Note this must be less than the kafka broker setting if we are using kafka
default_max_message_size = 14.5


class TransformerArgumentParser(argparse.ArgumentParser):

    def __init__(self, description="ServiceX Transformer"):
        super(TransformerArgumentParser, self).__init__(description=description)

        self.add_argument("--brokerlist", dest='brokerlist', action='store',
                          default=default_brokerlist,
                          help='List of Kafka broker to connect to')

        self.add_argument("--topic", dest='topic', action='store',
                          default='servicex',
                          help='Kafka topic to publish arrays to')

        self.add_argument("--chunks", dest='chunks', action='store',
                          default=None, type=int,
                          help='Arrow Buffer Chunksize')

        self.add_argument("--tree", dest='tree', action='store',
                          default="Events",
                          help='Tree from which columns will be inspected')

        self.add_argument("--attrs", dest='attr_names', action='store',
                          default=default_attr_names,
                          help='List of attributes to extract')

        self.add_argument("--path", dest='path', action='store',
                          default=None,
                          help='Path to single Root file to transform')

        self.add_argument("--limit", dest='limit', action='store',
                          default=None, type=int,
                          help='Max number of events to process')

        self.add_argument('--result-destination', dest='result_destination',
                          action='store',
                          default='kafka', help='kafka, object-store',
                          choices=['kafka', 'object-store', 'output-dir'])

        self.add_argument('--output-dir', dest='output_dir',
                          action='store',
                          default=None, help='Local directory to output results')

        self.add_argument('--result-format', dest='result_format', action='store',
                          default='arrow', help='arrow, parquet, root-file',
                          choices=['arrow', 'parquet', 'root-file'])

        self.add_argument("--max-message-size", dest='max_message_size',
                          action='store', default=default_max_message_size,
                          type=int,
                          help='Max message size in megabytes')

        self.add_argument('--rabbit-uri', dest="rabbit_uri", action='store',
                          default='host.docker.internal')

        self.add_argument('--request-id', dest='request_id', action='store',
                          default='servicex', help='Request ID to read from queue')

    @classmethod
    def extract_kafka_brokers(cls, brokerlist):
        return list(map(lambda b: b.strip(), brokerlist.split(",")))

    @classmethod
    def extract_attr_list(cls, attr_names):
        return list(map(lambda b: b.strip(), attr_names.split(",")))



# Copyright (c) 2019, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='servicex-transformer',
    packages=setuptools.find_packages(),
    version='0.2.4',
    license='bsd 3 clause',
    description='ServiceX Data Transformer for HEP Data',
    long_description=long_description,
    long_description_content_type="text/markdown",

    author='Ben Galewsky',
    author_email='bengal1@illinois.edu',
    url='https://github.com/ssl-hep/ServiceX_transformer',
    keywords=['HEP', 'Data Engineering', 'Data Lake'],
    install_requires=[
        'uproot',
        'awkward >= 0.12.0',
        'xxhash',
        'lz4',
        'requests >= 2.22.0',
        'pyarrow',
        'kafka',
        'confluent_kafka == 1.2.0',
        'pympler',
        'pika',
        'minio '
    ],

    extras_require={
        'test': ['flake8==3.5',
                 'coverage==4.5.2',
                 'codecov==2.0.15'],
    },

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
    ],
    python_requires='>=2.7',

)
